/******************************************************************************

    main.c

    Test demo program for USB multiple device access with USB-HUB.

    Test hardware is SBDBT32, it has PIC32MX695F512H.

	(C) S.Suwa 2013/5/12               http://www.suwa-koubou.jp   

******************************************************************************/

#include <stdio.h>
#include <string.h>
#include "GenericTypeDefs.h"
#include "HardwareProfile.h"
#include "usb.h"
#include "usb_host_uvc.h"
#include "usb_host_hid_parser.h"
#include "usb_host_hid.h"
#include "usb_host_msd.h"
#include "usb_mouse.h"
#include "FSIO.h"
#include "uart2.h"


/*------------------------------------------------------------------------------
	Configuration for PIC32MX (SBDBT32) 
------------------------------------------------------------------------------*/

#include "p32mx695f_config.h"


/*------------------------------------------------------------------------------
	definition
------------------------------------------------------------------------------*/

#define	putstr(x)	UART2PrintString(x)
#define	inkey()		UART2IsPressed()
#define	getch()		UART2GetChar()
#define	puthex(x)	UART2PutHex(x)
#define	putch(x)	UART2PutChar(x)
#define	putdec(x)	UART2PutDec(x)

typedef enum
{
	CAMERA_DETACHED = 0,
	CAMERA_ATTACHED,
    MOUSE_DETACHED,
    MOUSE_ATTACHED,
	MSD_DETACHED,
	MSD_ATTACHED,
} DEVICE_STATE;


/*------------------------------------------------------------------------------
	function prototype
------------------------------------------------------------------------------*/

void	InitializeSystem (void);
int     main (void);
void	putdata(BYTE *data, BYTE len, WORD type);
void    capture_store_image(void);
void	getstr(char *buf, BYTE length);
void	mouse_test(void);
void	memory_disk_test(void);


/*------------------------------------------------------------------------------
	global variable
------------------------------------------------------------------------------*/

// device state
DEVICE_STATE camera_state, mouse_state, msd_state;

// mouse
BYTE           mouse_address, msd_address;
USB_MOUSE_DATA moused;


/*------------------------------------------------------------------------
	InitializeSystem()
-----------------------------------------------------------------------*/

void InitializeSystem ( void )
{
	int	value;

	// Set System Config
	//SYSTEMConfigPerformance(GetSystemClock());

	value = SYSTEMConfigWaitStatesAndPB( GetSystemClock() );

	// Enable the cache for the best performance	
	CheKseg0CacheOn();

	INTEnableSystemMultiVectoredInt();

	value = OSCCON;
	while (!(value & 0x00000020)){
		value = OSCCON;    // Wait for PLL lock to stabilize	
	}

	// interrupt enable
	INTEnableInterrupts();

	// ADC input off
	AD1PCFG = 0xffff;

	// set UART port
	TRISD = 0xffff;
	TRISDCLR = 0x0008; // to digital OUT for UART1 TX
	LATDSET = 0x0008; // default level high for UART1 TX

	// SBDBT32 set LED port 
	TRISB = 0xffff;
	TRISBCLR = 0xc000;	// RB14,RB15 are digital OUT.
	LATBSET = 0xc000;	// RB14(LED1), RB15(LED2) are turn off.

    // Init UART
    UART2Init();	// include UART1

	// USB Host driver
	USBHostInit(0);


} // InitializeSystem



/*==============================================================================
	main
===============================================================================*/

int main ( void )
{
	int		c, n, m;
	char	buf[80];

	// UVC webcam
	BYTE	rc;
	BYTE	control, len;
	BYTE	data[64];
	WORD	type;
	BYTE 	maxformat, format, maxframe;
	UVC_FRAME_INFO	*frame_info;
	WORD 	width, height;


    // Initialize the processor and peripherals.
    InitializeSystem();

	// initialize status
	camera_state = CAMERA_DETACHED;
    mouse_state = MOUSE_DETACHED;
	msd_state = MSD_DETACHED;

	for(;;){
restart:
		// opening message
		putstr("\r\n*** USB multiple device accessing demo program *****\r\n");
		putstr("  wait for webcamera & mouse & memory disk attached.\r\n");
		while(camera_state == CAMERA_DETACHED || mouse_state == MOUSE_DETACHED || ! USBHostMSDSCSIMediaDetect())
        {
			USBTasks();
		}

		// clear key buffer
		while(inkey()) getch();

		// File System initialize
		if(!FSInit()){
			putstr("File system initialization error, in memory disk.\r\n");
            continue;
        }

		SetClockVars(2013, 5, 12, 9, 0, 0);

		for(;;){
			putstr("\r\n*** Demonstration program Menu *********************\r\n");
			putstr("m:Mouse read test.\r\n");
			putstr("d:Memory disk test.\r\n");
			putstr("v:List camera control current value.(Wheel Button)\r\n");
			putstr("i:List camera supported image format and image resolution.(Left Button)\r\n");
			putstr("c:Capture camera image data and write it to memory disk.(Right Button)\r\n"); 
			putstr("=>");

			// wait for typed key or pushed mouse button. 
            do {
                while((rc = mouse_read(mouse_address, &moused)) == FALSE && inkey() == 0){
                    USBTasks();
                    if(camera_state == CAMERA_DETACHED || mouse_state == MOUSE_DETACHED || ! USBHostMSDSCSIMediaDetect() ){
                        goto restart;
                    }
                }

                c = 0;
                if(rc == TRUE){
                    if(moused.button.wheel){
					    c = 'v';
                    }else if(moused.button.left){
                        c = 'i';
                    }else if(moused.button.right){
                        c = 'c';
                    }
                }else if(inkey()){
					c = getch();
				}

            } while(c == 0);


			if(c >= ' ' && c <= 0x7e){
				putch(c);
				putstr("\r\n");
			}

			switch(c){

			// mouse read test.
			case 'm':
				mouse_test();
				break;

			// memory disk test.
			case 'd':
				memory_disk_test();
				break;

			// List camera control value.
			case 'v':
				for(control = UVC_APL_SCANNING; control < UVC_APL_PROBE; control++){

					// first request is GET_CUR
					rc = USBHostUVCGetControlValue(control, GET_CUR, data, &len, &type);
					// if control isn't supported, loop continue to next control. 
					if(rc != UVC_SUCCESS) continue;

					// print control name
					putstr(USBHostUVCGetControlName(control)); putstr(":\r\n");
					// print current setting
					putstr("\tCur "); putdata(data, len, type); putstr("\r\n");

					// 2nd request is GET_MIN
					rc = USBHostUVCGetControlValue(control, GET_MIN, data, &len, &type);
					// if GET_MIN request isn't supported, loop continue to next control. 
					if(rc != UVC_SUCCESS) continue;

					// print minimum setting
					putstr("\tMin "); putdata(data, len, type); putstr("\r\n");

					// 3rd request is GET_MAX
					rc = USBHostUVCGetControlValue(control, GET_MAX, data, &len, &type);
					// if GET_MAX request isn't supported, loop continue to next control. 
					if(rc != UVC_SUCCESS) continue;

					// print maximum setting
					putstr("\tMax "); putdata(data, len, type); putstr("\r\n");

				} // for(control)
				break;

			// List camera supported image format and image resolution.
			case 'i':
				maxformat = USBHostUVCGetFormatCount();
				for(n = 0; n < maxformat; n++){
					format = USBHostUVCGetFormat(n);
					sprintf(buf, "Supported Format (Number is %d): ", n+1); putstr(buf);
					if(format == FMT_YUY2) putstr("YUY2");
					if(format == FMT_NV12) putstr("NV12");
					if(format == FMT_MJPEG) putstr("MJPEG");
					if(format == FMT_MPEG2TS) putstr("MPEG2 TS");
					if(format == FMT_DV) putstr("DV");
					putstr("\r\n");

					frame_info = USBHostUVCGetFrame(n, &maxframe);
					putstr("  Resolution:\r\n\t");
					for(m = 0; m < maxframe; m++){
						width = frame_info[m].wWidth;
						height = frame_info[m].wHeight;
						sprintf(buf, "%dx%d ", width, height); putstr(buf);
						if(m % 3 == 2) putstr("\r\n\t");
					}
					putstr("\r\n");
				}
				break;

			// Capture camera image data and write it to memory disk.
			case 'c':
				// check support MJPEG
				maxformat = USBHostUVCGetFormatCount();
				for(n = 0; n < maxformat; n++){
					if(USBHostUVCGetFormat(n) == FMT_MJPEG) break;
				}
				if(n >= maxformat){
					putstr(" Can't capture image, This camera doesn't support MJPEG image format.\r\n");
					break;
				}

				// First requested resolution is VGA.
				width = 640; height = 480;
				rc = USBHostUVCSetFrame(n, width, height);
				if(rc == UVC_UNSUPPORTED_RESOLUTION){
					// next resolution is QVGA.
					width = 320; height = 240;
					rc = USBHostUVCSetFrame(n, width, height);
					if(rc == UVC_UNSUPPORTED_RESOLUTION){
						// 3rd resolution is QQVGA.
						width = 160; height = 120;
						rc = USBHostUVCSetFrame(n, width, height);
						if(rc == UVC_UNSUPPORTED_RESOLUTION){
							putstr(" Can't capture image, This camera doesn't support VGA, QVGA, QQVGA resolution.\r\n");
							break;
						}
					}
				}
				if(rc != UVC_SUCCESS){
					putstr("Set Error!\r\n");
					break;
				}

				sprintf(buf, "Capture image format is MJPEG and resolution is %dx%d.\r\n", width, height);
				putstr(buf); 

				// Capture and write image.
				capture_store_image();

				// check device attached
				if(camera_state == CAMERA_DETACHED || mouse_state == MOUSE_DETACHED || ! USBHostMSDSCSIMediaDetect() ){
					goto restart;
				}
				break;


			default:
				break;

			} // switch(c)

		} // for (menu loop)

	}// for (main loop)

    return 0;

} // main 


/*------------------------------------------------------------------------------
	putdata 
------------------------------------------------------------------------------*/

void	putdata(BYTE *data, BYTE len, WORD type)
{
	BYTE	n;
	int		i;
	char	pbuf[10];

	if(data == NULL || len == 0 || type == 0) return;

	n = 0;
	while(type){
		switch(type & 0xf){ // check low half byte
		case BOOLEAN:
			putstr(data[n]? "1 ": "0 ");
			n++;
			break;

		case bBITMAP:
			for(i = 7; i >= 0; i--){
				if(data[n] & (1<<i)) putch('1');
				else putch('0');
			}
			n++;
			break;

		case wBITMAP:
			for(i = 15; i >= 0; i--){
				if(data[n] & (1<<i)) putch('1');
				else putch('0');
			}
			n += 2;
			break;

		case bSIGNED:
			sprintf(pbuf, "%d ", (int)data[n]); putstr(pbuf);
			n++;
			break;

		case wSIGNED:
			sprintf(pbuf, "%d ", ((int)(data[n+1]) << 8) + data[n]); putstr(pbuf);
			n += 2;
			break;

		case dwSIGNED:
			sprintf(pbuf, "%ld",  
				((long)(data[n+3]) << 24) + ((long)(data[n+2]) << 16)+ ((long)(data[n+1]) << 8) + data[n]
				); putstr(pbuf);
			n += 4;
			break;

		case bNUMBER:
			sprintf(pbuf, "%u ", (WORD)data[n]); putstr(pbuf);
			n++;
			break;

		case wNUMBER:
			sprintf(pbuf, "%u ", ((WORD)(data[n+1]) << 8) + data[n]); putstr(pbuf);
			n += 2;
			break;

		case dwNUMBER:
			sprintf(pbuf, "%lu",  
				((DWORD)(data[n+3]) << 24) + ((DWORD)(data[n+2]) << 16)+ ((DWORD)(data[n+1]) << 8) + data[n]
				); putstr(pbuf);
			n += 4;
			break;

		default:
			break;
		}
		type >>= 4;	// left shift half byte
	}

}


/*------------------------------------------------------------------------
	Capture image and image data write to memory disk
-----------------------------------------------------------------------*/

BYTE	frameBuffer[100000l];
#define	HEADER_LEN	12

void	capture_store_image(void)
{
	FSFILE	*fout;
	BYTE	frameID;
	UVC_DATA_BUFFER	*dp;
	DWORD	len, count;


	// file open for image stored.
	if((fout = FSfopen("CAPTURE.DAT", "w")) == NULL){
		putstr("Can't open file.\r\n");
		return;
	}

	putstr("Capture Start!\r\n");

	// start frame capturing
	if(USBHostUVCStartFrameCapturing() == UVC_FAIL){
		FSfclose(fout);
		putstr(" ... found capturing start error!\r\n");
		return;
	}

	for(;;){
		// search EndOfFrame
		LATBCLR=0x4000;	// turn on LED1
		for(;;){
			// Get Image data, and error check. 
			if((dp = USBHostUVCGetFrameCapturing()) == NULL){
				goto error_exit;
			}
			dp->valid = 0;

			// end of frame ?
			if(dp->length > HEADER_LEN && (dp->buffer[1] & 2) == 2){
				frameID = (dp->buffer[1] & 1) ? 0: 1;
				break;
			}
		}

		// search next frame start
		for(;;){
			// Get Image data, and error check. 
			if((dp = USBHostUVCGetFrameCapturing()) == NULL){
				goto error_exit;
			}

			// new frame Start
			if(dp->length > HEADER_LEN && dp->buffer[HEADER_LEN] == 0xff && dp->buffer[HEADER_LEN+1] == 0xD8){
				break;
			}

			dp->valid = 0;
		}
 
		// one frame image data write to memory buffer
		count = 0;
		do {
			len = dp->length - HEADER_LEN;
			if((count + len) < 100000l){
				memcpy(frameBuffer + count, dp->buffer + HEADER_LEN, len); 
				count += len;
			}
			dp->valid = 0;
			if((dp = USBHostUVCGetFrameCapturing()) == NULL){
				goto error_exit;
			}
		}while((dp->buffer[1] & 1) == frameID);
		dp->valid = 0;

		LATBSET=0x4000;	// turn off LED1

		// valid image data ? 
		if(*(frameBuffer+count-2) == 0xff &&  *(frameBuffer+count-1) == 0xd9 ){
			// image data write to disk, monitor LED on
			LATBCLR=0x8000;	// turn on LED2
			FSfwrite(frameBuffer, 1, count, fout);
			LATBSET=0x8000;	// turn off LED2
		}

		// if input any keys, capture stop.
		if(inkey()){
			getch();
			break;
		}

		// if pushed mouse left button or right button, capture stop
		if(mouse_read(mouse_address, &moused)){
			if(moused.button.left || moused.button.right){
				break;
			}
		}

	} // for()

error_exit:
	FSfclose(fout);

	// Stop Frame Capture
	USBHostUVCStopFrameCapturing();

	// LEDS turn off
	LATBSET = 0xc000;

	putstr("Capture Stop!\r\n");

}


/*===============================================================================
	memory disk test
===============================================================================*/

void	memory_disk_test(void)
{
	int			c, n;
	char		fpath[20];
	SearchRec	finfo;
	char		pbuf[64];
	FSFILE		*fin, *fout;
	size_t		size;
	unsigned long address;

	for(;;){
		putstr("\r\n*** Memory Disk Test Menu ************************\r\n");
		putstr("m: make the directory in current directory.\r\n");
		putstr("r: delete the directory in current directory.\r\n");
		putstr("c: change the current directory.\r\n");
		putstr("l: list file in current directory.\r\n");
		putstr("w: text write to a file.\r\n");
		putstr("t: read text from a file.\r\n");
		putstr("v: file binary dump.\r\n");
		putstr("d: delete the file.\r\n");
		putstr("f: format a memory disk.\r\n");
		putstr("q: quit and return to main menu.\r\n");
		putstr("=>");

		while(inkey() == 0){
			USBTasks();
			if(!USBHostMSDSCSIMediaDetect()){
				putstr("Found memory disk detached.\r\n");
				putstr("Stop test!\r\n");
				return;
			}
		}

		c = getch();
		if(c >= 'a' && c <= 'z') putch(c);
		putstr("\r\n");

		switch(c){
		case 'q':
			// return to main menu
			return;

		case 'm':
			// make sub directory
			putstr("make directory:\r\n");
			putstr("  directory name =>"); getstr(fpath, 20);
			if(FSmkdir(fpath) == 0){
				putstr("Directory was created successfully\r\n");
			}else{
				putstr("Directory could not be created.\r\n");
			}
			break;

		case 'r':
			// remove directory
			putstr("remove directory:\r\n");
			putstr("  directory name =>"); getstr(fpath, 20);
			if(FSrmdir(fpath, FALSE) == 0){
				putstr("Directory was deleted successfully\r\n");
			}else{
				putstr("Directory could not be deleted.\r\n");
			}
			break;

		case 'c':
			// change directory
			putstr("change directory:\r\n");
			putstr("  directory name =>"); getstr(fpath, 20);
			if(FSchdir(fpath) == 0){
				putstr("Directory was changed successfully\r\n");
			}else{
				putstr("Directory could not be changed.\r\n");
			}
			break;


		case 'l':
			// file listing in current directory
			if(FindFirst("*.*", ATTR_MASK, &finfo) == 0){
				putstr("\r\n");
				do {
					if(finfo.attributes == ATTR_DIRECTORY){
						sprintf(pbuf, "%12s <DIR>\r\n", finfo.filename);
					}else if(finfo.attributes == ATTR_VOLUME){
						sprintf(pbuf, "%12s <VOL>\r\n", finfo.filename);
					}else{
						sprintf(pbuf, "%12s %10ld\r\n", finfo.filename, finfo.filesize);
					}
					putstr(pbuf);
				}while(FindNext(&finfo) == 0);
			}
			break;

		case 'w':
			// write file
			putstr("write file:\r\n");
			putstr("  file name =>"); getstr(fpath, 20);
			if((fout = FSfopen(fpath, "w")) == NULL){
				putstr("File could not be opened.\r\n");
				break;
			}
			putstr("If will you stop the write operations, Enter key hit only. \r\n");

			for(;;){
				putstr("text string =>"); getstr(pbuf, 64);
				size = strlen(pbuf);
				if(size == 0) break;

				if(FSfwrite(pbuf, 1, size, fout) != size || 
						FSfwrite("\r\n", 1, 2, fout) != 2){
					putstr("write error!\r\n");
					break;
				}
			}

			putstr(" ... write complete.\r\n");
			FSfclose(fout);
			break;

		case 't':
			// read file
			putstr("read file:\r\n");
			putstr("  file name =>"); getstr(fpath, 20);
			if((fin = FSfopen(fpath, "r")) == NULL){
				putstr("File could not be opened.\r\n");
				break;
			}

			while((size = FSfread(pbuf, 1, 62, fin)) != 0){  
				pbuf[size] = 0;
				putstr(pbuf);
			}

			putstr(" ... read complete.\r\n");
			FSfclose(fin);
			break;

		case 'v':
			// binary dump
			putstr("binary dump file:\r\n");
			putstr("  file name =>"); getstr(fpath, 20);
			if((fin = FSfopen(fpath, "r")) == NULL){
				putstr("File could not be opened.\r\n");
				break;
			}
			putstr("         +0 +1 +2 +3 +4 +5 +6 +7 +8 +9 +A +B +C +D +E +F  0123456789ABCDEF\r\n");
			address = 0;
			while((size = FSfread(pbuf, 1, 16, fin)) != 0){
				sprintf(pbuf+16, "%08X ", address); putstr(pbuf+16);
				for(n = 0; n < size; n++){
					puthex(pbuf[n]); putch(' ');
				}
				for(; n < 16; n++) putstr("   ");
				putch(' ');
				for(n = 0; n < size; n++){
					if(pbuf[n] >= ' ' && pbuf[n] <= 0x7e){
						putch(pbuf[n]);
					}else{
						putch('.');
					}
				}

				address += size;
				putstr("\r\n");

				if((address % 0x100) == 0){
					putstr("Hit enter to dump continue, other key stop!\r\n");
					while(!inkey());
					c = getch();
					if(c != 0x0d) break;
				}
			}
			putstr(" ... dump complete.\r\n");
			FSfclose(fin);
			break;

		case 'd': 
       		// delete file.
			putstr("delete file:\r\n");
			putstr("  file name =>"); getstr(fpath, 20);
			if(FSremove(fpath) == 0){
				putstr("File deleted.\r\n");
			}else{
				putstr("File was not deleted.\r\n");
			}
			break;
			
		case 'f':
			// format a disk. Just erase the FAT and root
			putstr("format a disk ? (y/n) \r\n");
			while(inkey() == 0);
			if(getch() != 'y') break;

			putstr("\r\n");
			if(FSformat(0, 1, NULL) == 0){
				putstr("Format was successful.\r\n");
			}else{
				putstr("Format was unsuccessful.\r\n");
			}
			break;

		default:
			break;

		} // switch()

	} // for()	

}


/*===============================================================================
	mouse test
===============================================================================*/

void	mouse_test(void)
{
	USB_MOUSE_DATA   moused;

	putstr("Start mouse test. \r\n");
	putstr(" Push any keys when you will stop the test.\r\n"); 

	for(;;){

		while( ! mouse_read(mouse_address, &moused) ){
			if(mouse_state == MOUSE_DETACHED || inkey()){
				putstr("Stop mouse test! \r\n");
				return;
			}
		}

		if(moused.button.left){
			putstr(" left button on.\r\n");
		}

		if(moused.button.right){
			putstr(" right button on. \r\n");
		}

		if(moused.button.wheel){
			putstr(" wheel button on. \r\n");
		}

		if(moused.x_axis > 0){
			if( moused.y_axis > 0){
				putstr(" move to right down.\r\n");
			}

			if( moused.y_axis == 0){
				putstr(" move to right.\r\n");
			}

			if( moused.y_axis < 0){
				putstr(" move to right up.\r\n");
			}
		}

		if(moused.x_axis == 0){
			if( moused.y_axis > 0){
				putstr(" move to down.\r\n");
			}

			if( moused.y_axis < 0){
				putstr(" move to up.\r\n");
			}
		}
		
		if(moused.x_axis < 0){
			if( moused.y_axis > 0){
				putstr(" move to left down.\r\n");
			}

			if( moused.y_axis == 0){
				putstr(" move to left.\r\n");
			}

			if( moused.y_axis < 0){
				putstr(" move to left up.\r\n");
			}
		}

		if(moused.wheel < 0){
			putstr(" wheel up.\r\n");
		}

		if(moused.wheel > 0){
			putstr(" wheel down.\r\n");
		}

	} // for()

}


/*------------------------------------------------------------------------------
	getstr
------------------------------------------------------------------------------*/

void	getstr(char *buf, BYTE length)
{
	int	c, len;

	len = 0;
	for(;;){
		USBTasks();
		if(inkey()){
			c = getch();
			if(c >= ' ' && c <= 0x7e && len < (length -1)){
				*buf++ = c;
				len++;
				putch(c);       // echo back
			}else if(c == '\r'){
				putstr("\r\n");	// echo back
				break;
			}else if(c == 0x08 || c == 0x7f){
				if(len){
					putch(0x08);
					buf--;
					len--;
				}
			}
		}
	}
	*buf = 0;
}

/*-------------------------------------------------------------------------------
	BOOL USB_ApplicationEventHandler ( BYTE address, USB_EVENT event, void *data, DWORD size )
-------------------------------------------------------------------------------*/

BOOL USB_ApplicationEventHandler ( BYTE address, USB_EVENT event, void *data, DWORD size )
{

    // Handle specific events.
    switch (event){
    case EVENT_UVC_ATTACH:
		camera_state = CAMERA_ATTACHED;
        putstr("camera attached.\r\n");
        putstr(" address="); puthex(address); putstr("\r\n");
		break;

    case EVENT_UVC_DETACH:
    	camera_state = CAMERA_DETACHED;
        putstr("camera detached.\r\n" );
		break;

    case EVENT_HID_ATTACH:
        mouse_state = MOUSE_ATTACHED;
        mouse_address = address;
        putstr("mouse attached.\r\n");
        putstr(" address="); puthex(address); putstr("\r\n");
		break;

    case EVENT_HID_DETACH:
        mouse_state = MOUSE_DETACHED;
        putstr("mouse detached.\r\n");
		break;

	case EVENT_MSD_ATTACH:
		msd_state = MSD_ATTACHED;
		msd_address = address;
        putstr("memory disk attached.\r\n");
        putstr(" address="); puthex(address); putstr("\r\n");
		break;

    case EVENT_MSD_DETACH:
        msd_state = MSD_DETACHED;
        putstr("memory disk detached.\r\n");
		break;


	case EVENT_OVERRIDE_CLIENT_DRIVER_SELECTION:
		return FALSE;
 
	case EVENT_VBUS_RELEASE_POWER:
    case EVENT_VBUS_REQUEST_POWER:
    case EVENT_UNSUPPORTED_DEVICE:
    case EVENT_CANNOT_ENUMERATE:
    case EVENT_CLIENT_INIT_ERROR:
    case EVENT_OUT_OF_MEMORY:
    case EVENT_UNSPECIFIED_ERROR:
        break;

    case EVENT_HID_RPT_DESC_PARSED:
    #ifdef APPL_COLLECT_PARSED_DATA
	    return(APPL_COLLECT_PARSED_DATA());
	#endif
	    break;
    default:
       break;
   }

   return TRUE;

} // USB_ApplicationEventHandler


/*-------------------------------------------------------------------------------
	BOOL USB_ApplicationDataEventHandler ( BYTE address, USB_EVENT event, void *data, DWORD size )
-------------------------------------------------------------------------------*/

BOOL USB_ApplicationDataEventHandler ( BYTE address, USB_EVENT event, void *data, DWORD size )
{
	return TRUE;
}


/**** end of main.c ************************************************************/
