/******************************************************************************

	usb_mouse.c

*******************************************************************************/

#include "GenericTypeDefs.h"
#include "HardwareProfile.h"
#include "usb_config.h"
#include "usb.h"
#include "usb_host_hid_parser.h"
#include "usb_host_hid.h"
#include "usb_mouse.h"

//#define	DEBUG_MODE
#ifdef DEBUG_MODE
   #define	putstr(x)	UART2PrintString(x)
   #define	inkey()		UART2IsPressed()
   #define	getch()		UART2GetChar()
   #define	puthex(x)	UART2PutHex(x)
   #define	putch(x)	UART2PutChar(x)
   #define  putdec(x)   UART2PutDec(x)

   #include "uart2.h"
#endif

// *****************************************************************************
// *****************************************************************************
// Constants
// *****************************************************************************
// *****************************************************************************

// We are taking Timer 3  to schedule input report transfers
// NOTE - The datasheet doesn't state this, but the timer does get reset to 0
// after a period register match.  So we don't have to worry about resetting
// the timer manually.

#define STOP_TIMER_IN_IDLE_MODE     0x2000
#define TIMER_SOURCE_INTERNAL       0x0000
#define TIMER_ON                    0x8000
#define GATED_TIME_DISABLED         0x0000
#define TIMER_16BIT_MODE            0x0000

#define TIMER_PRESCALER_1           0x0000
#define TIMER_PRESCALER_8           0x0010
#define TIMER_PRESCALER_64          0x0020
#define TIMER_PRESCALER_256         0x0030
#define TIMER_INTERRUPT_PRIORITY    0x0001



// *** structures **************************************************************
typedef enum _APP_STATE
{
    DEVICE_NOT_CONNECTED,
    DEVICE_CONNECTED, /* Device Enumerated  - Report Descriptor Parsed */
    READY_TO_TX_RX_REPORT,
    GET_INPUT_REPORT, /* perform operation on received report */
    INPUT_REPORT_PENDING,
    ERROR_REPORTED 
} APP_STATE;

typedef struct _HID_REPORT_BUFFER
{
    WORD  Report_ID;
    WORD  ReportSize;
//    BYTE* ReportData;
    BYTE  ReportData[16];	// max 16byte
    WORD  ReportPollRate;
}   HID_REPORT_BUFFER;

// *** Internal Function Prototypes ********************************************
BYTE App_DATA2ASCII(BYTE a);
void AppInitialize(void);
BOOL AppGetParsedReportDetails(void);
void App_Detect_Device(void);
void App_ProcessInputReport(void);
BOOL USB_HID_DataCollectionHandler(void);

// *** Macros ******************************************************************
#define MAX_ALLOWED_CURRENT             (500)         // Maximum power we can supply in mA
#define MINIMUM_POLL_INTERVAL           (0x0A)        // Minimum Polling rate for HID reports is 10ms
#define USAGE_PAGE_BUTTONS              (0x09)
#define USAGE_PAGE_GEN_DESKTOP          (0x01)
#define MAX_ERROR_COUNTER               (10)

// *** Global Variables ********************************************************
APP_STATE App_State_Mouse = DEVICE_NOT_CONNECTED;

HID_DATA_DETAILS Appl_Mouse_Buttons_Details;
HID_DATA_DETAILS Appl_XY_Axis_Details;
HID_DATA_DETAILS Appl_Wheel_Axis_Details; // SUWA

HID_REPORT_BUFFER  Appl_raw_report_buffer;

HID_USER_DATA_SIZE Appl_Button_report_buffer[3];
HID_USER_DATA_SIZE Appl_XY_report_buffer[2];
HID_USER_DATA_SIZE Appl_Wheel_report_buffer[1]; // SUWA

BYTE ErrorDriver;
BYTE ErrorCounter;
BYTE NumOfBytesRcvd;

BOOL ReportBufferUpdated;


/*============================================================================================
	mouse_read() 
============================================================================================*/

BOOL	mouse_read(BYTE deviceAddress, USB_MOUSE_DATA *data)
{
	USBTasks();

	// DetectDevice
  	if(!USBHostHIDDeviceDetect(deviceAddress)){
  		App_State_Mouse = DEVICE_NOT_CONNECTED;
  	}

	switch(App_State_Mouse){
	case DEVICE_NOT_CONNECTED:
		USBTasks();
		/* True if report descriptor is parsed with no error */
		if(USBHostHIDDeviceDetect(deviceAddress)){
			App_State_Mouse = DEVICE_CONNECTED;
		}
		break;

	case DEVICE_CONNECTED:
		App_State_Mouse = READY_TO_TX_RX_REPORT;
		break;

	case READY_TO_TX_RX_REPORT:
		if(!USBHostHIDDeviceDetect(deviceAddress)){
			App_State_Mouse = DEVICE_NOT_CONNECTED;
		}else{
			App_State_Mouse = GET_INPUT_REPORT;
		}
		break;

	case GET_INPUT_REPORT:
		if(USBHostHIDRead(deviceAddress, Appl_raw_report_buffer.Report_ID,0,
			Appl_raw_report_buffer.ReportSize, Appl_raw_report_buffer.ReportData)){
				/* Host may be busy/error -- keep trying */
		}else{
			App_State_Mouse = INPUT_REPORT_PENDING;
		}
		USBTasks();
        break;

	case INPUT_REPORT_PENDING:
		if(USBHostHIDTransferIsComplete(deviceAddress, &ErrorDriver,&NumOfBytesRcvd)){
			if(ErrorDriver ||(NumOfBytesRcvd != Appl_raw_report_buffer.ReportSize )){
				ErrorCounter++ ; 
				if(MAX_ERROR_COUNTER <= ErrorDriver){
					App_State_Mouse = ERROR_REPORTED;
				}else{
					App_State_Mouse = READY_TO_TX_RX_REPORT;
				}
			}else{
				ErrorCounter = 0; 
				ReportBufferUpdated = TRUE;
				App_State_Mouse = READY_TO_TX_RX_REPORT;

				App_ProcessInputReport();
				data->button.left = Appl_Button_report_buffer[0] & 1;
				data->button.right = Appl_Button_report_buffer[1] & 1;
				data->button.wheel = Appl_Button_report_buffer[2] & 1;
				data->x_axis = Appl_XY_report_buffer[0];
				data->y_axis = Appl_XY_report_buffer[1];
				data->wheel = Appl_Wheel_report_buffer[0];

				return 	TRUE;
            }
		}
        break;

	case ERROR_REPORTED:	// forever error loop
		break;

	default:
		break;
	}

	// ReportData isn't ready!
	return FALSE;
}


/****************************************************************************
    void App_ProcessInputReport(void)
***************************************************************************/

void App_ProcessInputReport(void)
{

   /* process input report received from device */
    USBHostHID_ApiImportData(Appl_raw_report_buffer.ReportData, Appl_raw_report_buffer.ReportSize
                          ,Appl_Button_report_buffer, &Appl_Mouse_Buttons_Details);
    USBHostHID_ApiImportData(Appl_raw_report_buffer.ReportData, Appl_raw_report_buffer.ReportSize
                          ,Appl_XY_report_buffer, &Appl_XY_Axis_Details);

	/*** SUWA ****/
    USBHostHID_ApiImportData(Appl_raw_report_buffer.ReportData, Appl_raw_report_buffer.ReportSize
                          ,Appl_Wheel_report_buffer, &Appl_Wheel_Axis_Details);
	/*************/

#ifdef DEBUG_MODE
{
	int i;
	putstr( "HID: Raw Report \r\n");
	putstr("ReportSize="); putdec(Appl_raw_report_buffer.ReportSize);
	for(i=0;i<(Appl_raw_report_buffer.ReportSize);i++){
		puthex(Appl_raw_report_buffer.ReportData[i]);
	}
    putstr("\r\n");
	putstr("Left  Bt : "); putdec(Appl_Button_report_buffer[0]);putstr("\r\n");
	putstr("Right Bt : "); putdec(Appl_Button_report_buffer[1]);putstr("\r\n");
	putstr("CenterBt : "); putdec(Appl_Button_report_buffer[2]);putstr("\r\n");
    
	putstr("X-Axis : "); putdec((char)Appl_XY_report_buffer[0]);putstr("\r\n");
	putstr("Y-Axis : "); putdec((char)Appl_XY_report_buffer[1]);putstr("\r\n");
	putstr("Z-Axis : "); putdec((char)Appl_Wheel_report_buffer[0]);putstr("\r\n");
}
#endif

}


/*----------------------------------------------------------------------------------
	USB_HID_DataCollectionHandler()

	macro in usb_config.h
	#define APPL_COLLECT_PARSED_DATA USB_HID_DataCollectionHandler

	This function was called in USB_ApplicationEventHandler function.
----------------------------------------------------------------------------------*/

BOOL USB_HID_DataCollectionHandler(void)
{
  BYTE NumOfReportItem = 0;
  BYTE i;
  USB_HID_ITEM_LIST* pitemListPtrs;
  USB_HID_DEVICE_RPT_INFO* pDeviceRptinfo;
  HID_REPORTITEM *reportItem;
  HID_USAGEITEM *hidUsageItem;
  BYTE usageIndex;
  BYTE reportIndex;

  BYTE	xy_flg, wheel_flg;	// SUWA
  xy_flg = wheel_flg = 0;	// SUWA
  reportIndex = 0;			// SUWA


  pDeviceRptinfo = USBHostHID_GetCurrentReportInfo(); // Get current Report Info pointer
  pitemListPtrs = USBHostHID_GetItemListPointers();   // Get pointer to list of item pointers

  BOOL status = FALSE;
   /* Find Report Item Index for Modifier Keys */
   /* Once report Item is located , extract information from data structures provided by the parser */
   NumOfReportItem = pDeviceRptinfo->reportItems;
   for(i=0;i<NumOfReportItem;i++)
    {
       reportItem = &pitemListPtrs->reportItemList[i];
       if((reportItem->reportType==hidReportInput) && (reportItem->dataModes == (HIDData_Variable|HIDData_Relative))&&
           (reportItem->globals.usagePage==USAGE_PAGE_GEN_DESKTOP))
        {

			if(xy_flg == 0){	// SUWA

      	     	/* We now know report item points to modifier keys */
        	   	/* Now make sure usage Min & Max are as per application */
            	usageIndex = reportItem->firstUsageItem;
            	hidUsageItem = &pitemListPtrs->usageItemList[usageIndex];

            	reportIndex = reportItem->globals.reportIndex;
            	Appl_XY_Axis_Details.reportLength = (pitemListPtrs->reportList[reportIndex].inputBits + 7)/8;
            	Appl_XY_Axis_Details.reportID = (BYTE)reportItem->globals.reportID;
            	Appl_XY_Axis_Details.bitOffset = (BYTE)reportItem->startBit;
            	Appl_XY_Axis_Details.bitLength = (BYTE)reportItem->globals.reportsize;
            	Appl_XY_Axis_Details.count=(BYTE)reportItem->globals.reportCount;
            	Appl_XY_Axis_Details.interfaceNum= USBHostHID_ApiGetCurrentInterfaceNum();

				/*** SUWA ***/
				xy_flg = 1;	// SUWA

            	if(Appl_XY_Axis_Details.count==3){	// include wheel
	            	Appl_XY_Axis_Details.count = 2;
	            	Appl_Wheel_Axis_Details.reportLength = Appl_XY_Axis_Details.reportLength;
	            	Appl_Wheel_Axis_Details.reportID = Appl_XY_Axis_Details.reportID;
	            	Appl_Wheel_Axis_Details.bitOffset = Appl_XY_Axis_Details.bitOffset + Appl_XY_Axis_Details.bitLength * Appl_XY_Axis_Details.count;
            		Appl_Wheel_Axis_Details.bitLength = Appl_XY_Axis_Details.bitLength;
            		Appl_Wheel_Axis_Details.count=1;
            		Appl_Wheel_Axis_Details.interfaceNum= Appl_XY_Axis_Details.interfaceNum;
					wheel_flg = 1;
				}
				continue;
			} // if(xy_flg)

			if(wheel_flg == 0){

	           	/* We now know report item points to modifier keys */
   		       	/* Now make sure usage Min & Max are as per application */
            	usageIndex = reportItem->firstUsageItem;
            	hidUsageItem = &pitemListPtrs->usageItemList[usageIndex];

            	reportIndex = reportItem->globals.reportIndex;
            	Appl_Wheel_Axis_Details.reportLength = (pitemListPtrs->reportList[reportIndex].inputBits + 7)/8;
            	Appl_Wheel_Axis_Details.reportID = (BYTE)reportItem->globals.reportID;
            	Appl_Wheel_Axis_Details.bitOffset = (BYTE)reportItem->startBit;
            	Appl_Wheel_Axis_Details.bitLength = (BYTE)reportItem->globals.reportsize;
            	Appl_Wheel_Axis_Details.count= 1;
            	Appl_Wheel_Axis_Details.interfaceNum= USBHostHID_ApiGetCurrentInterfaceNum();

				wheel_flg = 1;
			}
			/*********/

        }
        else if((reportItem->reportType==hidReportInput) && (reportItem->dataModes == HIDData_Variable)&&
           (reportItem->globals.usagePage==USAGE_PAGE_BUTTONS))
        {
           /* We now know report item points to modifier keys */
           /* Now make sure usage Min & Max are as per application */
            usageIndex = reportItem->firstUsageItem;
            hidUsageItem = &pitemListPtrs->usageItemList[usageIndex];

            reportIndex = reportItem->globals.reportIndex;
            Appl_Mouse_Buttons_Details.reportLength = (pitemListPtrs->reportList[reportIndex].inputBits + 7)/8;
            Appl_Mouse_Buttons_Details.reportID = (BYTE)reportItem->globals.reportID;
            Appl_Mouse_Buttons_Details.bitOffset = (BYTE)reportItem->startBit;
            Appl_Mouse_Buttons_Details.bitLength = (BYTE)reportItem->globals.reportsize;
            Appl_Mouse_Buttons_Details.count=(BYTE)reportItem->globals.reportCount;
            Appl_Mouse_Buttons_Details.interfaceNum= USBHostHID_ApiGetCurrentInterfaceNum();
        }
    }

   if(pDeviceRptinfo->reports == 1)
    {
        Appl_raw_report_buffer.Report_ID = 0;
        Appl_raw_report_buffer.ReportSize = (pitemListPtrs->reportList[reportIndex].inputBits + 7)/8;
		// Appl_raw_report_buffer.ReportData = (BYTE*)malloc(Appl_raw_report_buffer.ReportSize);
        Appl_raw_report_buffer.ReportPollRate = pDeviceRptinfo->reportPollingRate;
        status = TRUE;
    }

    return(status);
}

/*----------------------------------------------------------------------------------
	USB_ApplicationEventHandler() 

	This function must be exists, because macro defined in usb_config.h
@@@@@@#define USB_HOST_APP_EVENT_HANDLER USB_ApplicationEventHandler
	macro use usb_host.c
-----------------------------------------------------------------------------------*/

#if 0   // move to main.c

#ifndef DEBUG_MODE
	#define putstr(x)
#endif

BOOL USB_ApplicationEventHandler( BYTE address, USB_EVENT event, void *data, DWORD size )
{
    switch( (INT)event )
    {
        case EVENT_VBUS_REQUEST_POWER:
            // The data pointer points to a byte that represents the amount of power
            // requested in mA, divided by two.  If the device wants too much power,
            // we reject it.
            if (((USB_VBUS_POWER_EVENT_DATA*)data)->current <= (MAX_ALLOWED_CURRENT / 2))
            {
                return TRUE;
            }
            else
            {
                putstr( "\r\n***** USB Error - device requires too much current *****\r\n" );
            }
            break;

        case EVENT_VBUS_RELEASE_POWER:
            // Turn off Vbus power.
            // The PIC24F with the Explorer 16 cannot turn off Vbus through software.
            return TRUE;
            break;

        case EVENT_HUB_ATTACH:
            putstr( "\r\n***** USB Error - hubs are not supported *****\r\n" );
            return TRUE;
            break;

        case EVENT_UNSUPPORTED_DEVICE:
            putstr( "\r\n***** USB Error - device is not supported *****\r\n" );
            return TRUE;
            break;

        case EVENT_CANNOT_ENUMERATE:
            putstr( "\r\n***** USB Error - cannot enumerate device *****\r\n" );
            return TRUE;
            break;

        case EVENT_CLIENT_INIT_ERROR:
            putstr( "\r\n***** USB Error - client driver initialization error *****\r\n" );
            return TRUE;
            break;

        case EVENT_OUT_OF_MEMORY:
            putstr( "\r\n***** USB Error - out of heap memory *****\r\n" );
            return TRUE;
            break;

        case EVENT_UNSPECIFIED_ERROR:   // This should never be generated.
            putstr( "\r\n***** USB Error - unspecified *****\r\n" );
            return TRUE;
            break;

		case EVENT_HID_RPT_DESC_PARSED:
			 #ifdef APPL_COLLECT_PARSED_DATA
			     return(APPL_COLLECT_PARSED_DATA());
		     #else
				 return TRUE;
			 #endif
			break;

        default:
            break;
    }
    return FALSE;
}
#endif // if 0


/*** end of usb_mouse.c **********************************************************************/
